import logging
import yaml

from bonfire.utils import RepoFile, get_dupes, FatalError

log = logging.getLogger(__name__)


def _fetch_apps_file(config):
    rf = RepoFile.from_config(config["appsFile"])
    commit, content = rf.fetch()
    log.info(
        "loading commit '%s' of %s repo %s/%s at path '%s' for apps config",
        commit,
        rf.host,
        rf.org,
        rf.repo,
        rf.path,
    )
    fetched_apps = yaml.safe_load(content)

    if "apps" not in fetched_apps:
        raise FatalError("fetched apps file has no 'apps' key")

    app_names = [a["name"] for a in fetched_apps["apps"]]
    dupes = get_dupes(app_names)
    if dupes:
        raise FatalError("duplicate app names found in fetched apps file: {dupes}")

    return {a["name"]: a for a in fetched_apps["apps"]}


def _parse_apps_in_cfg(config):
    app_names = [a["name"] for a in config["apps"]]
    dupes = get_dupes(app_names)
    if dupes:
        raise FatalError("duplicate app names found in config: {dupes}")
    return {a["name"]: a for a in config["apps"]}


def get_local_apps(config, fetch_remote=True):
    # get any apps set directly in config
    config_apps = {}
    if "apps" in config:
        config_apps = _parse_apps_in_cfg(config)
        log.info("local app configuration overrides found for: %s", list(config_apps.keys()))

    if not fetch_remote:
        final_apps = config_apps
    else:
        # fetch apps from repo if appsFile is provided in config
        fetched_apps = {}
        if "appsFile" in config:
            log.info("local config has a remote 'appsFile' defined, fetching it...")
            fetched_apps = _fetch_apps_file(config)

        # override fetched apps with local apps if any were defined
        fetched_apps.update(config_apps)
        final_apps = fetched_apps

    return final_apps



import json
import copy
import logging
from urllib.parse import urlparse

from gql import gql
from gql import Client as GQLClient
from gql.transport.requests import RequestsHTTPTransport
from requests.auth import HTTPBasicAuth

import bonfire.config as conf

log = logging.getLogger(__name__)


ENVS_QUERY = gql(
    """
    {
      envs: environments_v1 {
        name
        parameters
        namespaces {
          name
        }
      }
    }
    """
)

APPS_QUERY = gql(
    """
    {
      apps: apps_v1 {
        name
        parentApp {
          name
        }
        saasFiles {
          path
          name
          parameters
          resourceTemplates {
            name
            path
            url
            parameters
            targets {
              namespace {
                name
                cluster {
                  name
                }
              }
              ref
              parameters
            }
          }
        }
        saasFilesV2 {
          path
          name
          parameters
          resourceTemplates {
            name
            path
            url
            parameters
            targets {
              namespace {
                name
                cluster {
                  name
                }
              }
              ref
              parameters
            }
          }
        }
      }
    }
    """
)

NAMESPACE_QUERY = gql(
    """
    {
      namespaces: namespaces_v1 {
        name
        openshiftResources {
          ... on NamespaceOpenshiftResourceVaultSecret_v1 {
            name
            path
          }
        }
      }
    }
    """
)


class Client:
    def __init__(self):
        log.debug("using url: %s", conf.QONTRACT_BASE_URL)

        transport_kwargs = {"url": conf.QONTRACT_BASE_URL}

        if conf.QONTRACT_TOKEN:
            log.debug("using token authentication")
            transport_kwargs["headers"] = {"Authorization": conf.QONTRACT_TOKEN}
        elif conf.QONTRACT_USERNAME and conf.QONTRACT_PASSWORD:
            log.debug("using basic authentication")
            transport_kwargs["auth"] = HTTPBasicAuth(conf.QONTRACT_USERNAME, conf.QONTRACT_PASSWORD)

        transport = RequestsHTTPTransport(**transport_kwargs)
        self.client = GQLClient(transport=transport, fetch_schema_from_transport=True)

        # info level is way too noisy for the gql client
        logging.getLogger("gql").setLevel(logging.ERROR)

    def get_env(self, env):
        """Get insights env configuration."""
        for env_data in self.client.execute(ENVS_QUERY)["envs"]:
            if env_data["name"] == env:
                env_data["namespaces"] = set(n["name"] for n in env_data["namespaces"])
                break
        else:
            raise ValueError(f"cannot find env '{env}'")

        return env_data

    def get_apps(self):
        return self.client.execute(APPS_QUERY)["apps"]

    def get_namespace(self, name):
        for ns in self.client.execute(NAMESPACE_QUERY)["namespaces"]:
            if ns["name"] == name:
                return ns


_client = None


def get_client():
    global _client
    if not _client:
        _client = Client()

    return _client


def _to_dict(nullable_json_str):
    return json.loads(nullable_json_str or "{}")


def _find_matching_component(apps, app_name, component_name):
    for component in apps.get(app_name, {}).get("components", []):
        if component["name"] == component_name:
            return component


def _check_replace_other(other_params, this_params):
    this_clowder_enabled = bool(this_params.get("CLOWDER_ENABLED"))
    other_clowder_enabled = bool(other_params.get("CLOWDER_ENABLED"))
    if this_clowder_enabled and not other_clowder_enabled:
        return True

    this_replicas = int(this_params.get("REPLICAS", 1))
    other_replicas = int(other_params.get("REPLICAS", 1))
    if other_replicas < 1 and this_replicas > 1:
        return True

    this_replicas = int(this_params.get("MIN_REPLICAS", 1))
    other_replicas = int(other_params.get("MIN_REPLICAS", 1))
    if other_replicas < 1 and this_replicas > 1:
        return True

    return False


def _add_component_if_priority_higher(
    apps, app_name, component_name, env_name, saas_file, component, defined_multiple
):
    add_component = True
    existing_match = _find_matching_component(apps, app_name, component_name)

    if existing_match:
        # this app/component is defined multiple times in the environment
        # look at the parameters set on it to decide which definition to prioritize
        defined_multiple.add((app_name, component_name))
        other_params = existing_match["parameters"]
        this_params = component["parameters"]
        add_component = _check_replace_other(other_params, this_params)
        if add_component:
            log.debug(
                "app: '%s' component: '%s' defined in saas file '%s' takes priority",
                app_name,
                component_name,
                saas_file["path"],
            )
        else:
            log.debug(
                "app: '%s' component: '%s' defined in saas file '%s' has lower priority, skipped",
                app_name,
                component_name,
                saas_file["path"],
            )

    if add_component:
        apps[app_name]["name"] = app_name
        apps[app_name]["components"].append(component)
        log.debug(
            "app: '%s' component: '%s' added for env '%s' using saas file: %s",
            app_name,
            component_name,
            env_name,
            saas_file["name"],
        )


def _add_component(apps, env, app_name, saas_file, resource_template, target, defined_multiple):
    component_name = resource_template["name"]
    env_name = env["name"]
    saas_file_path = saas_file["path"]

    if app_name not in apps:
        apps[app_name] = {"name": app_name, "components": []}

    url = resource_template["url"]
    if "github" not in url and "gitlab" not in url:
        raise ValueError(
            f"unknown host for resourceTemplate url '{url}' found in saas file '{saas_file_path}'"
        )
    host = "github" if "github" in url else "gitlab"

    try:
        parsed_url = urlparse(url)
        org, repo = parsed_url.path.rstrip("/").split("/")[-2:]
    except (ValueError, IndexError) as err:
        raise ValueError(f"invalid repo url '{url}': {err}")

    # merge the various layers of parameters
    p = copy.deepcopy(_to_dict(env["parameters"]))
    p.update(_to_dict(saas_file["parameters"]))
    p.update(_to_dict(resource_template["parameters"]))
    p.update(_to_dict(target["parameters"]))

    component = {
        "name": component_name,
        "path": resource_template["path"],
        "host": host,
        "repo": f"{org}/{repo}",
        "ref": target["ref"],
        "parameters": p,
    }

    _add_component_if_priority_higher(
        apps, app_name, component_name, env_name, saas_file, component, defined_multiple
    )


def get_apps_for_env(env_name):
    client = get_client()
    all_apps = client.get_apps()
    env = client.get_env(env_name)

    # work-around to only show apps with an ephemeral deploy target
    if env_name == conf.EPHEMERAL_ENV_NAME:
        env["namespaces"] = [conf.BASE_NAMESPACE_NAME]

    apps = {}
    ignored_apps = set()
    defined_multiple = set()

    for app in all_apps:
        if app["parentApp"] and app["parentApp"].get("name") != "insights":
            ignored_apps.add(app["name"])
            continue
        saas_files = app.get("saasFiles", []) + app.get("saasFilesV2", [])
        for saas_file in saas_files:
            for resource_template in saas_file.get("resourceTemplates", []):
                for target in resource_template.get("targets", []):
                    if target["namespace"]["name"] in env["namespaces"]:
                        # this target belongs to the environment
                        _add_component(
                            apps,
                            env,
                            app["name"],
                            saas_file,
                            resource_template,
                            target,
                            defined_multiple,
                        )

    if ignored_apps:
        log.debug(
            "ignored apps in env '%s' that do not have parentApp 'insights': %s",
            env_name,
            ", ".join(ignored_apps),
        )

    if defined_multiple:
        log.debug(
            "the following components in env '%s' are defined multiple times: %s",
            env_name,
            ", ".join([f"{app}/{component}" for app, component in defined_multiple]),
        )

    return apps


def sub_refs(apps, ref_env_name):
    ref_env_apps = get_apps_for_env(ref_env_name)

    final_apps = copy.deepcopy(apps)
    for app_name, app in apps.items():
        for idx, component in enumerate(app["components"]):
            component_name = component["name"]
            ref_component = _find_matching_component(ref_env_apps, app_name, component_name)

            if ref_component:
                final_component = final_apps[app_name]["components"][idx]
                final_component["ref"] = ref_component["ref"]
                ref_image_tag = ref_component.get("parameters", {}).get("IMAGE_TAG")
                if ref_image_tag:
                    if "parameters" not in final_component:
                        final_component["parameters"] = {}
                    final_component["parameters"]["IMAGE_TAG"] = ref_image_tag
                log.debug(
                    "app: '%s' component: '%s' -- using ref from env '%s': %s%s",
                    app_name,
                    component_name,
                    ref_env_name,
                    final_component["ref"],
                    f", IMAGE_TAG: {ref_image_tag}" if ref_image_tag else "",
                )
            else:
                log.debug(
                    "app: '%s' component: '%s' -- no deploy cfg for env '%s', using ref 'master'",
                    app_name,
                    component_name,
                    ref_env_name,
                )
                final_apps[app_name]["components"][idx]["ref"] = "master"

    return final_apps


def get_namespaces_for_env(environment_name):
    client = get_client()

    namespaces = client.get_env(environment_name)["namespaces"]
    results = list(namespaces)
    log.debug("namespaces listed in qontract for environment '%s': %s", environment_name, results)
    return results


def get_secret_names_in_namespace(namespace_name):
    client = get_client()

    secret_names = []
    namespace = client.get_namespace(namespace_name)
    for resource in namespace["openshiftResources"]:
        if not resource:
            # query returns {} if resource is not 'NamespaceOpenshiftResourceVaultSecret_v1'
            continue
        name = resource["name"] or resource["path"].split("/")[-1]
        secret_names.append(name)
    return secret_names



import functools
import json
import logging
import re
import threading
import time

import sh
from sh import ErrorReturnCode, TimeoutException
from subprocess import PIPE
from subprocess import Popen

from ocviapy import export
from wait_for import wait_for, TimedOutError

log = logging.getLogger(__name__)


# assume that the result of this will not change during execution of our app
@functools.lru_cache(maxsize=None, typed=False)
def get_api_resources():
    output = oc("api-resources", verbs="list", _silent=True).strip()
    if not output:
        return []

    lines = output.split("\n")
    # lines[0] is the table header, use it to figure out length of each column
    groups = re.findall(r"(\w+\s+)", lines[0])

    name_start = 0
    name_end = len(groups[0])
    shortnames_start = name_end
    shortnames_end = name_end + len(groups[1])
    apigroup_start = shortnames_end
    apigroup_end = shortnames_end + len(groups[2])
    namespaced_start = apigroup_end
    namespaced_end = apigroup_end + len(groups[3])
    kind_start = namespaced_end

    resources = []
    for line in lines[1:]:
        shortnames = line[shortnames_start:shortnames_end].strip()
        resource = {
            "name": line[name_start:name_end].strip().rstrip("s") or None,
            "shortnames": shortnames.split(",") if shortnames else [],
            "apigroup": line[apigroup_start:apigroup_end].strip() or None,
            "namespaced": line[namespaced_start:namespaced_end].strip() == "true",
            "kind": line[kind_start:].strip() or None,
        }
        resources.append(resource)
    return resources


def parse_restype(string):
    """
    Given a resource type or its shortcut, return the full resource type name.
    """
    s = string.lower()
    for r in get_api_resources():
        if s in r["shortnames"] or s == r["name"]:
            return r["name"]

    raise ValueError("Unknown resource type: {}".format(string))


def _only_immutable_errors(err_lines):
    return all("field is immutable after creation" in line.lower() for line in err_lines)


def _conflicts_found(err_lines):
    return any("error from server (conflict)" in line.lower() for line in err_lines)


def _get_logging_args(args, kwargs):
    # Format the cmd args/kwargs for log printing before the command is run
    cmd_args = " ".join([str(arg) for arg in args if arg is not None])

    cmd_kwargs = []
    for key, val in kwargs.items():
        if key.startswith("_"):
            continue
        if len(key) > 1:
            cmd_kwargs.append("--{} {}".format(key, val))
        else:
            cmd_kwargs.append("-{} {}".format(key, val))
    cmd_kwargs = " ".join(cmd_kwargs)

    return cmd_args, cmd_kwargs


def _exec_oc(*args, **kwargs):
    _silent = kwargs.pop("_silent", False)
    _ignore_immutable = kwargs.pop("_ignore_immutable", True)
    _retry_conflicts = kwargs.pop("_retry_conflicts", True)
    _stdout_log_prefix = kwargs.pop("_stdout_log_prefix", " |stdout| ")
    _stderr_log_prefix = kwargs.pop("_stderr_log_prefix", " |stderr| ")

    kwargs["_bg"] = True
    kwargs["_bg_exc"] = False

    err_lines = []
    out_lines = []

    def _err_line_handler(line, _, process):
        threading.current_thread().name = f"pid-{process.pid}"
        if not _silent:
            log.info("%s%s", _stderr_log_prefix, line.rstrip())
        err_lines.append(line)

    def _out_line_handler(line, _, process):
        threading.current_thread().name = f"pid-{process.pid}"
        if not _silent:
            log.info("%s%s", _stdout_log_prefix, line.rstrip())
        out_lines.append(line)

    retries = 3
    last_err = None
    for count in range(1, retries + 1):
        cmd = sh.oc(*args, **kwargs, _tee=True, _out=_out_line_handler, _err=_err_line_handler)
        if not _silent:
            cmd_args, cmd_kwargs = _get_logging_args(args, kwargs)
            log.info("running (pid %d): oc %s %s", cmd.pid, cmd_args, cmd_kwargs)
        try:
            return cmd.wait()
        except ErrorReturnCode as err:
            # Sometimes stdout/stderr is empty in the exception even though we appended
            # data in the callback. Perhaps buffers are not being flushed ... so just
            # set the out lines/err lines we captured on the Exception before re-raising it by
            # re-init'ing the err and causing it to rebuild its message template.
            #
            # see https://github.com/amoffat/sh/blob/master/sh.py#L381
            err.__init__(
                full_cmd=err.full_cmd,
                stdout="\n".join(out_lines).encode(),
                stderr="\n".join(err_lines).encode(),
                truncate=err.truncate,
            )

            # Make these plain strings for easier exception handling
            err.stdout = "\n".join(out_lines)
            err.stderr = "\n".join(err_lines)

            last_err = err
            # Ignore warnings that are printed to stderr in our error analysis
            err_lines = [line for line in err_lines if not line.lstrip().startswith("Warning:")]

            # Check if these are errors we should handle
            if _ignore_immutable and _only_immutable_errors(err_lines):
                log.warning("Ignoring immutable field errors")
                break
            elif _retry_conflicts and _conflicts_found(err_lines):
                log.warning(
                    "Hit resource conflict, retrying in 1 sec (attempt %d/%d)",
                    count,
                    retries,
                )
                time.sleep(1)
                continue

            # Bail if not
            raise last_err
    else:
        log.error("Retried %d times, giving up", retries)
        raise last_err


def oc(*args, **kwargs):
    """
    Run 'sh.oc' and print the command, show output, catch errors, etc.

    Optional kwargs:
        _ignore_errors: if ErrorReturnCode is hit, don't re-raise it (default False)
        _silent: don't print command or resulting stdout (default False)
        _ignore_immutable: ignore errors related to immutable objects (default True)
        _retry_conflicts: retry commands if a conflict error is hit
        _stdout_log_prefix: prefix this string to stdout log output (default " |stdout| ")
        _stderr_log_prefix: prefix this string to stderr log output (default " |stderr| ")

    Returns:
        None if cmd fails and _exit_on_err is False
        command output (str) if command succeeds
    """
    _ignore_errors = kwargs.pop("_ignore_errors", False)
    # The _silent/_ignore_immutable/_retry_conflicts kwargs are passed on so don't pop them yet

    try:
        return _exec_oc(*args, **kwargs)
    except ErrorReturnCode:
        if not _ignore_errors:
            raise
        else:
            if not kwargs.get("_silent"):
                log.warning("Non-zero return code ignored")


# we will assume that 'oc whoami' will not change during execution
@functools.lru_cache(maxsize=None, typed=False)
def whoami():
    name = oc("whoami", _silent=True).strip()
    # a valid label must be an empty string or consist of alphanumeric characters,
    # '-', '_' or '.', and must start and end with an alphanumeric character, so let's just sanitize
    # the name at this point
    return name.replace("@", "_at_").replace(":", "_")


def apply_config(namespace, list_resource):
    """
    Apply a k8s List of items
    """
    if namespace is None:
        oc("apply", "-f", "-", _in=json.dumps(list_resource))
    else:
        oc("apply", "-f", "-", "-n", namespace, _in=json.dumps(list_resource))


def get_json(restype, name=None, label=None, namespace=None):
    """
    Run 'oc get' for a given resource type/name/label and return the json output.

    If name is None all resources of this type are returned

    If label is not provided, then "oc get" will not be filtered on label
    """
    restype = parse_restype(restype)

    args = ["get", restype]
    if name:
        args.append(name)
    if label:
        args.extend(["-l", label])
    if namespace:
        args.extend(["-n", namespace])
    try:
        output = oc(*args, o="json", _silent=True)
    except ErrorReturnCode as err:
        if "NotFound" in err.stderr:
            return {}
        raise

    try:
        parsed_json = json.loads(str(output))
    except ValueError:
        return {}

    return parsed_json


def get_routes(namespace):
    """
    Get all routes in the project.

    Return dict with key of service name, value of http route
    """
    data = get_json("route", namespace=namespace)
    ret = {}
    for route in data.get("items", []):
        ret[route["metadata"]["name"]] = route["spec"]["host"]
    return ret


class StatusError(Exception):
    pass


# resources we are able to parse the status of
_CHECKABLE_RESOURCES = [
    "deploymentconfig",
    "deployment",
    "statefulset",
    "daemonset",
    "clowdapp",
    "clowdenvironment",
    "kafka",
    "kafkaconnect",
    "pod",
]


def _is_checkable(kind):
    return kind.lower() in _CHECKABLE_RESOURCES


def _available_checkable_resources(namespaced=False):
    """Returns resources we are able to parse status of that are present on the cluster."""
    if namespaced:
        return [
            r["kind"].lower()
            for r in get_api_resources()
            if _is_checkable(r["kind"]) and r["namespaced"]
        ]

    return [r["kind"].lower() for r in get_api_resources() if _is_checkable(r["kind"])]


def _resources_for_ns_wait():
    """Only check "higher level" resource types when waiting on a namespace"""
    resources = _available_checkable_resources(namespaced=True)
    try:
        resources.remove("pod")
    except ValueError as err:
        if "not in list" in str(err):
            pass
        else:
            raise
    return resources


def _get_name_for_kind(kind):
    for r in get_api_resources():
        if r["kind"].lower() == kind.lower():
            return r["name"]
    raise ValueError(f"unable to find resource name for kind '{kind}'")


def _check_status_for_restype(restype, json_data):
    """
    Depending on the resource type, check that it is "ready" or "complete"

    Uses the status json from an 'oc get'

    Returns True if ready, False if not.
    """
    restype = parse_restype(restype)

    if restype != "pod" and restype not in _CHECKABLE_RESOURCES:
        raise ValueError(f"Checking status for resource type {restype} currently not supported")

    try:
        status = json_data["status"]
    except KeyError:
        status = None

    if not status:
        return False

    generation = json_data["metadata"].get("generation")
    status_generation = status.get("observedGeneration") or status.get("generation")
    if generation and status_generation and generation != status_generation:
        return False

    if restype == "deploymentconfig" or restype == "deployment":
        spec_replicas = json_data["spec"]["replicas"]
        available_replicas = status.get("availableReplicas", 0)
        updated_replicas = status.get("updatedReplicas", 0)
        if available_replicas == spec_replicas and updated_replicas == spec_replicas:
            return True

    elif restype == "statefulset":
        spec_replicas = json_data["spec"]["replicas"]
        ready_replicas = status.get("readyReplicas", 0)
        return ready_replicas == spec_replicas

    elif restype == "daemonset":
        desired = status.get("desiredNumberScheduled", 1)
        available = status.get("numberAvailable")
        return desired == available

    elif restype == "pod":
        if status.get("phase").lower() == "running":
            return True

    elif restype in ("clowdenvironment", "clowdapp"):
        return str(status.get("ready")).lower() == "true"

    elif restype in ("kafka", "kafkaconnect"):
        conditions = status.get("conditions", [])
        for c in conditions:
            if str(c.get("status")).lower() == "true" and c.get("type").lower() == "ready":
                return True


def _get_resource_info(item):
    kind = item["kind"].lower()
    restype = _get_name_for_kind(kind)
    name = item["metadata"]["name"]
    key = f"{restype}/{name}"
    return kind, restype, name, key


class ResourceWaiter:
    def __init__(self, namespace, restype, name):
        self.namespace = namespace
        self.restype = parse_restype(restype)
        self.name = name.lower()
        self.observed_resources = dict()
        self._uid = None
        self.key = f"{self.restype}/{self.name}"
        self._time_last_logged = None
        self._time_remaining = None

        if self.restype not in _available_checkable_resources():
            raise ValueError(
                f"unable to check status of '{self.restype}' resources on this cluster"
            )

    def _observe(self, item):
        _, restype, _, key = _get_resource_info(item)
        if key not in self.observed_resources:
            self.observed_resources[key] = {"ready": False}
        if not self.observed_resources[key]["ready"]:
            if _check_status_for_restype(restype, item):
                log.info("[%s] resource is ready!", key)
                self.observed_resources[key]["ready"] = True

    def check_ready(self):
        response = get_json(self.restype, name=self.name, namespace=self.namespace)
        if response:
            self._uid = response["metadata"]["uid"]
            self._observe(response)
            return all([r["ready"] is True for _, r in self.observed_resources.items()])
        return False

    def _check_with_periodic_log(self):
        if self.check_ready():
            return True

        if time.time() > self._time_last_logged + 60:
            self._time_remaining -= 60
            if self._time_remaining:
                log.info("[%s] waiting %dsec longer", self.key, self._time_remaining)
                self._time_last_logged = time.time()
        return False

    def wait_for_ready(self, timeout, reraise=False):
        self._time_last_logged = time.time()
        self._time_remaining = timeout

        try:
            # check for ready initially, only wait_for if we need to
            log.debug("[%s] checking if 'ready'", self.key)
            if not self.check_ready():
                log.info("[%s] waiting up to %dsec for resource to be 'ready'", self.key, timeout)
                wait_for(
                    self._check_with_periodic_log,
                    message=f"wait for {self.key} to be 'ready'",
                    delay=5,
                    timeout=timeout,
                )
            return True
        except (StatusError, ErrorReturnCode) as err:
            log.error("[%s] hit error waiting for resource to be ready: %s", self.key, str(err))
            if reraise:
                raise
        except (TimeoutException, TimedOutError):
            log.error("[%s] timed out waiting for resource to be ready", self.key)
            if reraise:
                raise
        return False


class ResourceOwnerWaiter(ResourceWaiter):
    def _update_observed_resources(self, item):
        for owner_ref in item["metadata"].get("ownerReferences", []):
            restype_matches = owner_ref["kind"].lower() == self.restype
            owner_uid_matches = owner_ref["uid"] == self._uid
            if restype_matches and owner_uid_matches:
                _, restype, _, resource_key = _get_resource_info(item)
                if resource_key not in self.observed_resources:
                    self.observed_resources[resource_key] = {"ready": False}
                    log.info(
                        "[%s] found owned resource %s",
                        self.key,
                        resource_key,
                    )

                # check if ready state has transitioned for this resource
                if not self.observed_resources[resource_key]["ready"]:
                    if _check_status_for_restype(restype, item):
                        log.info("[%s] owned resource %s is ready!", self.key, resource_key)
                        self.observed_resources[resource_key]["ready"] = True

    def _observe(self, item):
        super()._observe(item)
        for restype in _available_checkable_resources():
            response = get_json(restype, namespace=self.namespace)
            for item in response.get("items", []):
                self._update_observed_resources(item)


def wait_for_ready(namespace, restype, name, timeout=300):
    waiter = ResourceWaiter(namespace, restype, name)
    return waiter.wait_for_ready(timeout)


def wait_for_ready_threaded(waiters, timeout=300):
    threads = [
        threading.Thread(target=waiter.wait_for_ready, daemon=True, args=(timeout,))
        for waiter in waiters
    ]
    for thread in threads:
        thread.name = thread.name.lower()
        thread.start()
    for thread in threads:
        thread.join()

    all_failed_resources = set()
    for waiter in waiters:
        waiter_failed_resources = [
            key for key, val in waiter.observed_resources.items() if val["ready"] is False
        ]
        for failed_resource in waiter_failed_resources:
            all_failed_resources.add(failed_resource)

    if all_failed_resources:
        log.info("some resources failed to become ready: %s", ", ".join(all_failed_resources))
        return False
    return True


def _all_resources_ready(namespace, timeout):
    already_waited_on = set()

    # wait on ClowdEnvironment, if there's one using this ns as its targetNamespace
    start = time.time()

    clowd_env = find_clowd_env_for_ns(namespace)
    if clowd_env:
        waiter = ResourceOwnerWaiter(namespace, "clowdenvironment", clowd_env["metadata"]["name"])
        if not waiter.wait_for_ready(timeout):
            return False

        for key in waiter.observed_resources:
            already_waited_on.add(key)

    end = time.time()
    elapsed = end - start
    timeout = int(timeout - elapsed)

    # wait on all ClowdApps in this namespace
    start = time.time()

    waiters = []
    clowdapps = get_json("clowdapp", namespace=namespace)
    for clowdapp in clowdapps["items"]:
        waiter = ResourceOwnerWaiter(namespace, "clowdapp", clowdapp["metadata"]["name"])
        waiters.append(waiter)
    if not wait_for_ready_threaded(waiters, timeout):
        return False

    for waiter in waiters:
        for key in waiter.observed_resources:
            already_waited_on.add(key)

    end = time.time()
    elapsed = end - start
    timeout = int(timeout - elapsed)

    # wait on anything else not covered by the above
    waiters = []
    for restype in _resources_for_ns_wait():
        response = get_json(restype, namespace=namespace)
        for item in response.get("items", []):
            _, restype, name, resource_key = _get_resource_info(item)
            if resource_key not in already_waited_on:
                waiter = ResourceWaiter(namespace, restype, name)
                waiters.append(waiter)

    return wait_for_ready_threaded(waiters, timeout)


def wait_for_all_resources(namespace, timeout=300):
    # wrap the other wait_fors in 1 wait_for so overall timeout is honored
    # wait_for returns a tuple of the return code and the time taken
    wait_for(
        _all_resources_ready,
        func_args=(namespace, timeout),
        message="wait for all deployed resources to be ready",
        timeout=timeout,
    )


def wait_for_db_resources(namespace, timeout=300):
    clowdapps = get_json("clowdapp", namespace=namespace).get("items", [])
    if len(clowdapps) == 0:
        raise ValueError(f"no clowdapps found in ns '{namespace}', no DB's to wait for")

    waiters = []
    for clowdapp in clowdapps:
        clowdapp_name = clowdapp["metadata"]["name"]
        db_name = clowdapp["spec"].get("database", {}).get("name")
        if db_name:
            waiters.append(ResourceWaiter(namespace, "deployment", f"{clowdapp_name}-db"))
        shared_db_app_name = clowdapp["spec"].get("database", {}).get("sharedDbAppName")
        if shared_db_app_name:
            waiters.append(ResourceWaiter(namespace, "deployment", f"{shared_db_app_name}-db"))

    if not waiters:
        raise ValueError(
            f"no clowdapps with db configurations found in '{namespace}', no DB's to wait for"
        )

    wait_for_ready_threaded(waiters, timeout)


def copy_namespace_secrets(src_namespace, dst_namespace, secret_names):
    for secret_name in secret_names:
        secret_data = export("secret", secret_name, namespace=src_namespace)
        ignore = secret_data["metadata"].get("annotations", {}).get("bonfire.ignore")
        if str(ignore).lower() == "true":
            log.debug(
                "secret '%s' in namespace '%s' has bonfire.ignore==true, skipping",
                secret_name,
                src_namespace,
            )
            continue

        log.info(
            "copying secret '%s' from namespace '%s' to namespace '%s'",
            secret_name,
            src_namespace,
            dst_namespace,
        )
        oc(
            "apply",
            f="-",
            n=dst_namespace,
            _in=json.dumps(secret_data),
            _silent=True,
        )


def process_template(template_data, params):
    valid_pnames = set(p["name"] for p in template_data.get("parameters", []))
    param_str = " ".join(f"-p {k}='{v}'" for k, v in params.items() if k in valid_pnames)

    proc = Popen(
        f"oc process --local --ignore-unknown-parameters -o json -f - {param_str}",
        shell=True,
        stdin=PIPE,
        stdout=PIPE,
    )
    stdout, stderr = proc.communicate(json.dumps(template_data).encode("utf-8"))
    return json.loads(stdout.decode("utf-8"))


def find_clowd_env_for_ns(ns):
    try:
        clowd_envs = get_json("clowdenvironment")
    except ErrorReturnCode as err:
        log.debug("hit error running 'oc get clowdenvironment': %s", err)
        clowd_envs = {"items": []}

    for clowd_env in clowd_envs["items"]:
        target_ns = clowd_env["spec"].get("targetNamespace")
        # in case target ns was not defined in the spec, check the env's status...
        target_ns = target_ns or clowd_env.get("status", {}).get("targetNamespace")
        if target_ns == ns:
            return clowd_env


def get_clowd_env_target_ns(clowd_env_name):
    try:
        clowd_env = get_json("clowdenvironment", clowd_env_name)
    except ErrorReturnCode as err:
        log.debug("hit error running 'oc get clowdenvironment %s': %s", clowd_env_name, err)
        return None

    return clowd_env.get("status", {}).get("targetNamespace")


def wait_for_clowd_env_target_ns(clowd_env_name):
    log.info("waiting for Clowder to provision target namespace for env '%s'", clowd_env_name)
    return wait_for(
        get_clowd_env_target_ns,
        func_args=(clowd_env_name,),
        fail_condition=None,
        num_sec=60,
        message="wait for Clowder to provision target namespace",
    ).out


# assume that the result of this will not change during execution of a single 'bonfire' command
@functools.lru_cache(maxsize=None, typed=False)
def on_k8s():
    """Detect whether this is a k8s or openshift cluster based on existence of projects."""
    project_resource = [r for r in get_api_resources() if r["name"] == "project"]

    if project_resource:
        return False
    return True


def get_all_namespaces():
    if not on_k8s():
        all_namespaces = get_json("project")["items"]
    else:
        all_namespaces = get_json("namespace")["items"]

    return all_namespaces


def wait_on_cji(namespace, cji_name, timeout):
    # first wait for job associated with this CJI to appear
    log.info("waiting for Job to appear owned by CJI '%s'", cji_name)

    def _find_job():
        jobs = get_json("job", label=f"clowdjob={cji_name}", namespace=namespace)
        try:
            return jobs["items"][0]["metadata"]["name"]
        except (KeyError, IndexError):
            return False

    job_name, elapsed = wait_for(
        _find_job, num_sec=timeout, message=f"wait for Job to appear owned by CJI '{cji_name}'"
    )

    log.info(
        "found Job '%s' created by CJI '%s', now waiting for pod to appear", job_name, cji_name
    )

    def _pod_found():
        pods = get_json("pod", label=f"job-name={job_name}", namespace=namespace)
        try:
            return pods["items"][0]["metadata"]["name"]
        except (KeyError, IndexError):
            return False

    remaining_time = timeout - elapsed

    pod_name, elapsed = wait_for(
        _pod_found,
        num_sec=remaining_time,
        message=f"wait for Pod to appear owned by CJI '{cji_name}'",
    )

    log.info(
        "found pod '%s' associated with CJI '%s', now waiting for pod to be 'running'",
        pod_name,
        cji_name,
    )

    remaining_time = remaining_time - elapsed

    waiter = ResourceWaiter(namespace, "pod", pod_name)
    waiter.wait_for_ready(remaining_time, reraise=True)

    return pod_name


def wait_on_reservation(res_name, timeout):
    log.info("waiting for reservation '%s' to get picked up by operator", res_name)

    def _find_reservation():
        res = get_json("reservation", name=res_name)
        try:
            return res["status"]["namespace"]
        except (KeyError, IndexError):
            return False

    ns_name, elapsed = wait_for(
        _find_reservation,
        num_sec=timeout,
        message=f"waiting for namespace to be allocated to reservation '{res_name}'",
    )
    return ns_name


def check_for_existing_reservation(requester):
    log.info("Checking for existing reservations for '%s'", requester)

    all_res = get_json("reservation")

    for res in all_res["items"]:
        if res["spec"]["requester"] == requester:
            return True

    return False


def get_reservation(name=None, namespace=None, requester=None):
    if name:
        res = get_json("reservation", name=name)
        return res if res else False
    elif namespace:
        all_res = get_json("reservation")
        for res in all_res["items"]:
            if res["status"]["namespace"] == namespace:
                return res
    elif requester:
        all_res = get_json("reservation", label=f"requester={requester}")
        numRes = len(all_res["items"])
        if numRes == 0:
            return False
        elif numRes == 1:
            return all_res["items"][0]
        else:
            log.info("Multiple reservations found for requester '%s'. Aborting.", requester)
            return False

    return False



import copy
import datetime
import logging
import json
import random
import time
import uuid
import yaml
import threading
from wait_for import TimedOutError

import bonfire.config as conf
from bonfire.qontract import get_namespaces_for_env, get_secret_names_in_namespace
from bonfire.openshift import (
    oc,
    on_k8s,
    get_all_namespaces,
    get_json,
    copy_namespace_secrets,
    process_template,
    wait_for_all_resources,
    whoami,
)


NS_RESERVED = "ephemeral-ns-reserved"
NS_READY = "ephemeral-ns-ready"
NS_REQUESTER = "ephemeral-ns-requester"
NS_DURATION = "ephemeral-ns-duration"
NS_EXPIRES = "ephemeral-ns-expires"
NS_REQUESTER_NAME = "ephemeral-ns-requester-name"
NS_REQUIRED_LABELS = [
    NS_DURATION,
    NS_EXPIRES,
    NS_READY,
    NS_REQUESTER,
    NS_REQUESTER_NAME,
    NS_RESERVED,
]


RESERVATION_DELAY_SEC = 5

log = logging.getLogger(__name__)


TIME_FMT = "%Y-%m-%d_T%H-%M-%S_%Z"


def _utc_tz(dt):
    return dt.replace(tzinfo=datetime.timezone.utc)


def _parse_time(string):
    return _utc_tz(datetime.datetime.strptime(string, TIME_FMT)) if string else None


def _fmt_time(dt):
    return datetime.datetime.strftime(_utc_tz(dt), TIME_FMT) if dt else None


def _utcnow():
    return _utc_tz(datetime.datetime.utcnow())


def _pretty_time_delta(seconds):
    # https://gist.github.com/thatalextaylor/7408395
    seconds = int(seconds)
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if days > 0:
        return "%dd%dh%dm%ds" % (days, hours, minutes, seconds)
    elif hours > 0:
        return "%dh%dm%ds" % (hours, minutes, seconds)
    elif minutes > 0:
        return "%dm%ds" % (minutes, seconds)
    else:
        return "%ds" % (seconds,)


class Namespace:
    def refresh(self):
        self.__init__(namespace_data=get_json("namespace", self.name))
        return self

    def __init__(self, name=None, namespace_data=None):
        if not namespace_data:
            # we don't yet have the data for this namespace... load it
            if not name:
                raise ValueError('Namespace needs one of: "name", "namespace_data"')
            self.name = name
            self.refresh()
            return

        self.data = copy.deepcopy(namespace_data)
        self.name = self.data["metadata"]["name"]

        self._initialize_labels = False
        if "labels" not in self.data["metadata"]:
            self.data["metadata"]["labels"] = {}
            self._initialize_labels = True

        self.labels = self.data["metadata"]["labels"]

        self.reserved = self.labels.get(NS_RESERVED, "false") == "true"
        self.ready = self.labels.get(NS_READY, "false") == "true"
        requester = self.labels.get(NS_REQUESTER)
        self.requester = str(requester) if requester else None
        duration = self.labels.get(NS_DURATION)
        self.duration = int(duration) if duration else None
        # convert time format to one that can be used in a label
        self.expires = _parse_time(self.labels.get(NS_EXPIRES))
        requester_name = self.labels.get(NS_REQUESTER_NAME)
        self.requester_name = str(requester_name) if requester_name else None

    @property
    def is_reservable(self):
        """
        Check whether a namespace has the required labels set on it.
        """
        if all([label in self.labels for label in NS_REQUIRED_LABELS]):
            return True
        return False

    @property
    def expires_in(self):
        if not self.expires:
            # reconciler needs to set the time ...
            return "TBD"
        utcnow = _utcnow()
        delta = self.expires - utcnow
        return _pretty_time_delta(delta.total_seconds())

    @property
    def owned_by_me(self):
        if on_k8s():
            return True
        return self.requester_name == whoami()

    @property
    def available(self):
        return not self.reserved and self.ready

    def __str__(self):
        return (
            f"ns {self.name} (reservable: {self.is_reservable}, owned_by_me: {self.owned_by_me}, "
            f"available: {self.available})"
        )

    def update(self):
        patch = []

        if self._initialize_labels:
            # prevent 'The  "" is invalid' error due to missing 'labels' path
            patch.append({"op": "add", "path": "/metadata/labels", "value": {}})

        patch.extend(
            [
                {
                    "op": "replace",
                    "path": f"/metadata/labels/{NS_RESERVED}",
                    "value": str(self.reserved).lower(),
                },
                {
                    "op": "replace",
                    "path": f"/metadata/labels/{NS_READY}",
                    "value": str(self.ready).lower(),
                },
                {
                    "op": "replace",
                    "path": f"/metadata/labels/{NS_REQUESTER}",
                    "value": str(self.requester) if self.requester else None,
                },
                {
                    "op": "replace",
                    "path": f"/metadata/labels/{NS_DURATION}",
                    "value": str(self.duration) if self.duration else None,
                },
                {
                    "op": "replace",
                    "path": f"/metadata/labels/{NS_EXPIRES}",
                    # convert time format to one that can be used in a label
                    "value": _fmt_time(self.expires),
                },
                {
                    "op": "replace",
                    "path": f"/metadata/labels/{NS_REQUESTER_NAME}",
                    "value": str(self.requester_name) if self.requester_name else None,
                },
            ]
        )

        oc("patch", "namespace", self.name, type="json", p=json.dumps(patch))


def _get_env_ready_status():
    clowd_env_ready_for_ns = {}
    clowd_envs = get_json("clowdenvironment")
    for clowd_env in clowd_envs["items"]:
        status = clowd_env.get("status", {})
        target_ns = status.get("targetNamespace")
        ready = status.get("ready", False)
        clowd_env_ready_for_ns[target_ns] = ready
        if not ready:
            log.debug("found target ns '%s' with env status not ready", target_ns)
    return clowd_env_ready_for_ns


def get_namespaces(available=False, mine=False):
    """
    Look up reservable namespaces in the cluster.

    available (bool) -- return only namespaces that are ready and not reserved
    mine (bool) -- return only namespaces owned by current user
    """
    log.debug("get_namespaces(available=%s, mine=%s)", available, mine)
    all_namespaces = [Namespace(namespace_data=ns) for ns in get_all_namespaces()]

    log.debug("namespaces found:\n%s", "\n".join([str(n) for n in all_namespaces]))

    # get clowd envs to ensure that ClowdEnvironment is ready for the namespaces
    env_ready_for_ns = _get_env_ready_status()

    ephemeral_namespaces = []
    for ns in all_namespaces:
        ns.ready = ns.ready and env_ready_for_ns.get(ns.name, False)
        if not ns.is_reservable:
            continue
        get_all = not mine and not available
        if get_all or (mine and ns.owned_by_me) or (available and ns.available):
            ephemeral_namespaces.append(ns)

    return ephemeral_namespaces


def _reserve_ns_for_duration(namespace, duration):
    requester_id = uuid.uuid4()
    namespace.reserved = True
    namespace.ready = False
    namespace.requester = requester_id
    namespace.duration = duration
    namespace.expires = None  # let the reconciler tell us when it expires
    namespace.requester_name = whoami()
    namespace.update()
    return requester_id


def _should_renew_ns(namespace, duration):
    # if the ns is already reserved by us, don't re-reserve it if the requested duration is less
    # than the expires_in time (in other words, if you have reserved an ns already that expires in
    # 2 days you would not want to reset it to expire in 60 minutes)
    if namespace.owned_by_me:
        if not namespace.expires:
            # we're not sure when this expires yet, let's not re-reserve it
            log.warning("namespace owned by you but expires time unknown, not renewing")
            return False

        expires_in_delta = namespace.expires - _utcnow()
        expires_in_hrs = expires_in_delta.total_seconds() / 3600
        if duration <= expires_in_hrs:
            log.warning(
                "namespace owned by you, expires in '%s' >= duration '%dh', not renewing",
                _pretty_time_delta(expires_in_delta.total_seconds()),
                duration,
            )
            return False

    return True


def reserve_namespace(duration, retries, specific_namespace=None, attempt=0):
    attempt = attempt + 1

    ns_name = specific_namespace if specific_namespace else ""
    log.info("attempt [%d] to reserve namespace %s", attempt, ns_name)

    if specific_namespace:
        log.debug("specific namespace requested: %s", ns_name)
        # look up both available ns's and ns's owned by 'me' to allow for renewing reservation
        available_namespaces = get_namespaces(available=True, mine=True)
        available_namespaces = [ns for ns in available_namespaces if ns.name == specific_namespace]
    else:
        # if a specific namespace was not requested, only look up available ones
        available_namespaces = get_namespaces(available=True, mine=False)

    if not available_namespaces:
        log.info("all namespaces currently unavailable")

        if retries and attempt > retries:
            log.error("maximum retries reached")
            return None

        log.info("waiting 60sec before retrying")
        time.sleep(60)
        return reserve_namespace(duration, retries, specific_namespace, attempt=attempt)

    namespace = random.choice(available_namespaces)

    if not _should_renew_ns(namespace, duration):
        return namespace

    requester_id = _reserve_ns_for_duration(namespace, duration)

    # to avoid race conditions, wait and verify we still own this namespace
    time.sleep(RESERVATION_DELAY_SEC)
    namespace.refresh()
    if str(namespace.requester) != str(requester_id):
        log.warning("hit namespace reservation conflict")

        if retries and attempt > retries:
            log.error("maximum retries reached")
            return None

        return reserve_namespace(duration, retries, specific_namespace, attempt=attempt)

    return namespace


def release_namespace(namespace):
    ns = Namespace(name=namespace)
    ns.reserved = False
    ns.requester = None
    ns.requester_name = None
    ns.ready = False
    ns.update()


def _delete_resources(namespace):
    # delete some of our own operator resources first
    resources_to_delete = [
        "cyndipipelines",  # delete this first to prevent hanging
        "xjoinpipelines",
        "clowdjobinvocations",
        "clowdapps",
    ]
    for resource in resources_to_delete:
        oc("delete", resource, "--all", n=namespace, timeout="60s")

    # delete the ClowdEnvironment for this namespace
    if get_json("clowdenvironment", conf.ENV_NAME_FORMAT.format(namespace=namespace)):
        oc(
            "delete",
            "clowdenvironment",
            conf.ENV_NAME_FORMAT.format(namespace=namespace),
            timeout="60s",
        )

    # delete the FrontendEnvironment for this namespace
    if get_json("frontendenvironment", conf.ENV_NAME_FORMAT.format(namespace=namespace)):
        oc(
            "delete",
            "frontendenvironment",
            conf.ENV_NAME_FORMAT.format(namespace=namespace),
            timeout="60s",
        )

    # delete any other lingering specific resource types from the namespace
    resources_to_delete = [
        "ingresses",
        "frontends",
        "bundles",
        "elasticsearches",
        "horizontalpodautoscalers",
        "kafkabridges",
        "kafkaconnectors",
        "kafkaconnects",
        "kafkaconnects2is",
        "kafkamirrormaker2s",
        "kafkamirrormakers",
        "kafkarebalances",
        "kafkas",
        "kafkatopics",
        "kafkausers",
        "deployments",
        "deploymentconfigs",
        "statefulsets",
        "daemonsets",
        "replicasets",
        "cronjobs",
        "jobs",
        "services",
        "routes",
        "pods",
        "secrets",
        "configmaps",
        "persistentvolumeclaims",
    ]
    for resource in resources_to_delete:
        oc("delete", resource, "--all", n=namespace, timeout="60s")


def add_base_resources(namespace, secret_names):
    copy_namespace_secrets(conf.BASE_NAMESPACE_NAME, namespace, secret_names)

    with open(conf.EPHEMERAL_CLUSTER_CLOWDENV_TEMPLATE) as fp:
        template_data = yaml.safe_load(fp)

    processed_template = process_template(
        template_data,
        params={
            "ENV_NAME": conf.ENV_NAME_FORMAT.format(namespace=namespace),
            "NAMESPACE": namespace,
        },
    )

    oc("apply", f="-", _in=json.dumps(processed_template))

    # wait for any deployed base resources to become 'ready'
    wait_for_all_resources(namespace, timeout=conf.RECONCILE_TIMEOUT)


def _reconcile_ns(ns, base_secret_names):
    log.info("namespace '%s' - checking", ns.name)
    update_needed = False

    if ns.reserved and ns.expires:
        # check if the reservation has expired
        utcnow = _utcnow()
        log.info("namespace '%s' - expires: %s, utcnow: %s", ns.name, ns.expires, utcnow)
        if utcnow > ns.expires:
            log.info("namespace '%s' - reservation expired, releasing", ns.name)
            ns.reserved = False
            ns.ready = False
            ns.duration = None
            ns.expires = None
            ns.requester = None
            ns.requester_name = None
            _delete_resources(ns.name)
            update_needed = True
        log.info("namespace '%s' - not expired", ns.name)

    if not ns.reserved and not ns.ready:
        # check if any released namespaces need to be prepped
        log.info("namespace '%s' - not reserved, but not 'ready', prepping", ns.name)
        _delete_resources(ns.name)
        try:
            add_base_resources(ns.name, base_secret_names)
        except TimedOutError:
            # base resources failed to come up, don't mark it ready and try again next time...
            log.error("namespace '%s' - timed out waiting for resources after prep", ns.name)
        else:
            ns.ready = True
            ns.duration = None
            ns.expires = None
            ns.requester = None
            ns.requester_name = None
            update_needed = True

    if ns.reserved and ns.duration and not ns.expires:
        # this is a newly reserved namespace, set the expires time
        log.info("namespace '%s' - setting expiration time", ns.name)
        ns.expires = _utcnow() + datetime.timedelta(hours=ns.duration)
        update_needed = True

    if update_needed:
        ns.update()

    log.info("namespace '%s' - done", ns.name)


def get_namespaces_for_reconciler():
    """
    Query app-interface to get list of namespaces the reconciler operates on.
    """
    ephemeral_namespace_names = get_namespaces_for_env(conf.EPHEMERAL_ENV_NAME)
    log.debug(
        "namespaces found for env '%s': %s", conf.EPHEMERAL_ENV_NAME, ephemeral_namespace_names
    )
    all_namespaces = get_json("project")["items"]
    log.debug(
        "all namespaces found on cluster: %s", [ns["metadata"]["name"] for ns in all_namespaces]
    )
    ephemeral_namespaces = []

    # get clowd envs to ensure that ClowdEnvironment is ready for the namespaces
    env_ready_for_ns = _get_env_ready_status()
    for ns in all_namespaces:
        ns_name = ns["metadata"]["name"]
        if ns_name == conf.BASE_NAMESPACE_NAME:
            log.debug("ns '%s' is base namespace, will not reconcile it")
            continue
        if ns_name not in ephemeral_namespace_names:
            log.debug(
                "ns '%s' is not a member of env '%s', will not reconcile it",
                ns_name,
                conf.EPHEMERAL_ENV_NAME,
            )
            continue
        if not conf.RESERVABLE_NAMESPACE_REGEX.match(ns_name):
            log.debug(
                "ns '%s' does not match reservable namespace regex, will not reconcile it", ns_name
            )
            continue
        ns = Namespace(namespace_data=ns)
        ns.ready = ns.ready and env_ready_for_ns.get(ns.name, False)
        ephemeral_namespaces.append(ns)

    return ephemeral_namespaces


def reconcile():
    # run graphql queries outside of the threads since the client isn't natively thread-safe
    namespaces = get_namespaces_for_reconciler()
    base_secret_names = get_secret_names_in_namespace(conf.BASE_NAMESPACE_NAME)

    threads = []
    for ns in namespaces:
        t = threading.Thread(target=_reconcile_ns, args=(ns, base_secret_names))
        t.name = ns.name
        threads.append(t)
        t.start()
    for t in threads:
        t.join()



import copy
import logging
import json
import yaml
import re
import traceback
import uuid

from pathlib import Path

import bonfire.config as conf
from bonfire.openshift import process_template, whoami
from bonfire.utils import RepoFile, FatalError
from bonfire.utils import get_dependencies as utils_get_dependencies

log = logging.getLogger(__name__)


def _remove_resource_config(items):
    for i in items:
        if i["kind"] != "ClowdApp":
            continue

        removed = False
        for d in i["spec"].get("deployments", []):
            if "resources" in d["podSpec"]:
                del d["podSpec"]["resources"]
                removed = True
        for p in i["spec"].get("pods", []):
            if "resources" in p:
                del p["resources"]
                removed = True

        if removed:
            log.debug("removed resources from ClowdApp '%s'", i["metadata"]["name"])


def _set_replicas(items):
    for i in items:
        if i["kind"] != "ClowdApp":
            continue

        updated = False
        for d in i["spec"].get("deployments", []):
            if "minReplicas" in d["podSpec"] and d["podSpec"]["minReplicas"] > 1:
                d["podSpec"]["minReplicas"] = 1
                updated = True
        for p in i["spec"].get("pods", []):
            if "minReplicas" in p and p["minReplicas"] > 1:
                p["minReplicas"] = 1
                updated = True

        if updated:
            log.debug("set replicas to '1' on ClowdApp '%s'", i["metadata"]["name"])


def process_clowd_env(target_ns, quay_user, env_name, template_path):
    log.info("processing ClowdEnvironment")

    env_template_path = Path(template_path if template_path else conf.DEFAULT_CLOWDENV_TEMPLATE)

    if not env_template_path.exists():
        raise FatalError("ClowdEnvironment template file does not exist: %s", env_template_path)

    with env_template_path.open() as fp:
        template_data = yaml.safe_load(fp)

    params = dict()
    params["ENV_NAME"] = env_name
    if quay_user:
        quay_user = quay_user.replace("_", "-")
        params["PULL_SECRET_NAME"] = f"{quay_user}-pull-secret"
    if target_ns:
        params["NAMESPACE"] = target_ns

    processed_template = process_template(template_data, params=params)

    if not processed_template.get("items"):
        raise FatalError("Processed ClowdEnvironment template has no items")

    return processed_template


def process_iqe_cji(
    clowd_app_name,
    debug=False,
    marker="",
    filter="",
    env="clowder_smoke",
    image_tag="",
    cji_name=None,
    template_path=None,
    requirements="",
    requirements_priority="",
    test_importance="",
):
    log.info("processing IQE ClowdJobInvocation")

    template_path = Path(template_path if template_path else conf.DEFAULT_IQE_CJI_TEMPLATE)

    if not template_path.exists():
        raise FatalError("CJI template file does not exist: %s", template_path)

    with template_path.open() as fp:
        template_data = yaml.safe_load(fp)

    requirements = requirements.split(",") if requirements else []
    requirements_priority = requirements_priority.split(",") if requirements_priority else []
    test_importance = test_importance.split(",") if test_importance else []

    params = dict()
    params["DEBUG"] = str(debug).lower()
    params["MARKER"] = marker
    params["FILTER"] = filter
    params["ENV_NAME"] = env
    params["IMAGE_TAG"] = image_tag
    params["NAME"] = cji_name or f"iqe-{str(uuid.uuid4()).split('-')[0]}"
    params["APP_NAME"] = clowd_app_name
    params["REQUIREMENTS"] = json.dumps(requirements)
    params["REQUIREMENTS_PRIORITY"] = json.dumps(requirements_priority)
    params["TEST_IMPORTANCE"] = json.dumps(test_importance)

    processed_template = process_template(template_data, params=params)

    if not processed_template.get("items"):
        raise FatalError("Processed CJI template has no items")

    return processed_template


def process_reservation(
    name,
    requester,
    duration,
    template_path=None,
):
    log.info("processing namespace reservation")

    template_path = Path(template_path if template_path else conf.DEFAULT_RESERVATION_TEMPLATE)

    if not template_path.exists():
        raise FatalError("Reservation template file does not exist: %s", template_path)

    with template_path.open() as fp:
        template_data = yaml.safe_load(fp)

    params = dict()

    params["NAME"] = name if name else f"bonfire-reservation-{str(uuid.uuid4()).split('-')[0]}"
    params["DURATION"] = duration

    if requester is None:
        try:
            requester = whoami()
        except Exception:
            log.info("whoami returned an error - setting requester to 'bonfire'")  # minikube
            requester = "bonfire"

    params["REQUESTER"] = requester

    processed_template = process_template(template_data, params=params)

    if not processed_template.get("items"):
        raise FatalError("Processed Reservation template has no items")

    return processed_template


class TemplateProcessor:
    @staticmethod
    def _parse_app_names(app_names):
        parsed_app_names = set()
        for app_name in app_names:
            # backward compatibility for specifying app names with comma,separated,values
            for entry in app_name.split(","):
                parsed_app_names.add(entry)
        return parsed_app_names

    @staticmethod
    def _find_dupe_components(components_for_app):
        """Make sure no component is listed more than once across all apps."""
        for app_name, components in components_for_app.items():
            components_for_other_apps = copy.copy(components_for_app)
            del components_for_other_apps[app_name]

            for component in components:
                found_in = [app_name]
                for other_app_name, other_components in components_for_other_apps.items():
                    if component in other_components:
                        found_in.append(other_app_name)
                if len(found_in) > 1:
                    raise FatalError(
                        f"component '{component}' is not unique, found in apps: {found_in}"
                    )

    @staticmethod
    def _validate_component_dict(all_components, data, name):
        # Translate deprecated "APP_NAME/COMPONENT_NAME" path syntax
        updated_data = {}

        # Max number of items we expect separated by '/' in the CLI option value
        max_len = 2
        if name == "--set-parameter":
            max_len = 3

        for path, value in data.items():
            split = path.split("/")
            if len(split) == max_len:
                # first item was an app name
                new_path = split[1:]
            elif len(split) == max_len - 1:
                # first item was a component name
                new_path = split[0:]
            else:
                raise FatalError(f"invalid format for {name}: {path}={value}")

            component_name = new_path[0]

            # Make sure component name actually exists in app config
            if component_name not in all_components:
                raise FatalError(
                    f"component given for {name} not found in app config: {component_name}"
                )

            key = "/".join(new_path)
            updated_data[key] = value

        # Update the paths
        data.clear()
        for key, val in updated_data.items():
            data[key] = val

    @staticmethod
    def _validate_component_list(all_components, data, name):
        for component_name in data:
            if component_name not in all_components:
                raise FatalError(
                    f"component given for {name} not found in app config: {component_name}"
                )

    def _validate_component_options(self, all_components, data, name):
        if isinstance(data, dict):
            self._validate_component_dict(all_components, data, name)
        else:
            self._validate_component_list(all_components, data, name)

    def _validate(self):
        """
        Validate app configurations and options passed to the TemplateProcessor

        1. Check that each app has required keys
        2. Check that each app name is unique
        3. Check that each component in an app has required keys
        4. Check that each component is a unique name across the whole config
        5. Check that CLI params requiring a component use a valid component name
        """
        components_for_app = {}

        for app_name, app_cfg in self.apps_config.items():
            # Check that each app has required keys
            required_keys = ["name", "components"]
            missing_keys = [k for k in required_keys if k not in app_cfg]
            if missing_keys:
                raise FatalError(f"app '{app_name}' is missing required keys: {missing_keys}")

            # Check that each app name is unique
            app_name = app_cfg["name"]
            if app_name in components_for_app:
                raise FatalError(f"app with name '{app_name}' is not unique")
            components_for_app[app_name] = []

            for component in app_cfg.get("components", []):
                # Check that each component in an app has required keys
                required_keys = ["name", "host", "repo", "path"]
                missing_keys = [k for k in required_keys if k not in component]
                if missing_keys:
                    raise FatalError(
                        f"component on app {app_name} is missing required keys: {missing_keys}"
                    )
                comp_name = component["name"]
                components_for_app[app_name].append(comp_name)

        # Check that each component name is unique across the whole config
        self._find_dupe_components(components_for_app)

        # Check that CLI params requiring a component use a valid component name
        all_components = []
        for _, app_components in components_for_app.items():
            all_components.extend(app_components)

        log.debug("components found: %s", all_components)

        self._validate_component_options(
            all_components, self.template_ref_overrides, "--set-template-ref"
        )
        self._validate_component_options(all_components, self.param_overrides, "--set-parameter")

        # 'all' is a valid component keyword for these options below
        all_components.append("all")

        self._validate_component_options(
            all_components, self.remove_resources, "--remove-resources"
        )
        self._validate_component_options(
            all_components, self.no_remove_resources, "--no-remove-resources"
        )
        self._validate_component_options(all_components, self.component_filter, "--component")

    def __init__(
        self,
        apps_config,
        app_names,
        get_dependencies,
        image_tag_overrides,
        template_ref_overrides,
        param_overrides,
        clowd_env,
        remove_resources,
        no_remove_resources,
        single_replicas,
        component_filter,
    ):
        self.apps_config = apps_config
        self.requested_app_names = self._parse_app_names(app_names)
        self.get_dependencies = get_dependencies
        self.image_tag_overrides = image_tag_overrides
        self.template_ref_overrides = template_ref_overrides
        self.param_overrides = param_overrides
        self.clowd_env = clowd_env
        self.remove_resources = remove_resources
        self.no_remove_resources = no_remove_resources
        self.single_replicas = single_replicas
        self.component_filter = component_filter

        self._validate()

        self.k8s_list = {
            "kind": "List",
            "apiVersion": "v1",
            "metadata": {},
            "items": [],
        }

        self.processed_components = set()

    def _get_app_config(self, app_name):
        if app_name not in self.apps_config:
            raise FatalError(f"app {app_name} not found in apps config")
        return self.apps_config[app_name]

    def _get_component_config(self, component_name):
        for _, app_cfg in self.apps_config.items():
            for component in app_cfg["components"]:
                if component["name"] == component_name:
                    return component
        else:
            raise FatalError(f"component with name '{component_name}' not found")

    def _sub_image_tags(self, items):
        content = json.dumps(items)
        for image, image_tag in self.image_tag_overrides.items():
            # easier to just re.sub on a whole string
            content, subs = re.subn(rf"{image}:[-\w\.]+", rf"{image}:{image_tag}", content)
            if subs:
                log.info("replaced %d occurence(s) of image tag for image '%s'", subs, image)
        return json.loads(content)

    def _sub_ref(self, current_component_name, repo_file):
        for component_name, value in self.template_ref_overrides.items():
            if current_component_name == component_name:
                log.info(
                    "component: '%s' overriding template ref to '%s'",
                    component_name,
                    value,
                )
                repo_file.ref = value

    def _sub_params(self, current_component_name, params):
        for path, value in self.param_overrides.items():
            split = path.split("/")
            if len(split) == 2:
                component_name, param_name = split
            else:
                raise FatalError(f"invalid format for --set-parameter: {path}={value}")

            if current_component_name == component_name:
                log.info(
                    "component: '%s' overriding param '%s' to '%s'",
                    component_name,
                    param_name,
                    value,
                )
                params[param_name] = value

    def _get_component_items(self, component_name):
        component = self._get_component_config(component_name)
        try:
            rf = RepoFile.from_config(component)
            # override template ref if requested
            self._sub_ref(component_name, rf)
            commit, template_content = rf.fetch()
        except Exception as err:
            log.error("failed to fetch template file for %s", component_name)
            log.debug(traceback.format_exc())
            raise FatalError(err)

        template = yaml.safe_load(template_content)

        params = {
            "IMAGE_TAG": commit[:7],
            "ENV_NAME": self.clowd_env,
        }

        params.update(component.get("parameters", {}))

        # override any specific parameters on this component if requested
        self._sub_params(component_name, params)

        new_items = process_template(template, params)["items"]

        # override the tags for all occurences of an image if requested
        new_items = self._sub_image_tags(new_items)

        if (
            "all" not in self.no_remove_resources
            and ("all" in self.remove_resources or component_name in self.remove_resources)
            and component_name not in self.no_remove_resources
        ):
            _remove_resource_config(new_items)
        if self.single_replicas:
            _set_replicas(new_items)

        return new_items

    def _process_component(self, component_name):
        if component_name not in self.processed_components:
            log.info("processing component %s", component_name)
            new_items = self._get_component_items(component_name)
            self.k8s_list["items"].extend(new_items)

            self.processed_components.add(component_name)

            if self.get_dependencies:
                # recursively process components to add config for dependent apps to self.k8s_list
                self._add_dependencies_to_config(component_name, new_items)
        else:
            log.debug("component %s already processed", component_name)

    def _add_dependencies_to_config(self, component_name, new_items):
        dependencies = set()
        for _, deps in utils_get_dependencies(new_items).items():
            dependencies = dependencies.union(deps)

        # filter out ones we've already processed before
        dependencies = [d for d in dependencies if d not in self.processed_components]

        if dependencies:
            log.info("dependencies not previously processed: %s", dependencies)
            for component_name in dependencies:
                self._process_component(component_name)

    def _process_app(self, app_name):
        log.info("processing app '%s'", app_name)
        app_cfg = self._get_app_config(app_name)
        for component in app_cfg["components"]:
            component_name = component["name"]
            log.debug("app '%s' has component '%s'", app_name, component_name)
            if self.component_filter and component_name not in self.component_filter:
                log.debug(
                    "skipping component '%s', not found in --component filter", component_name
                )
                continue
            self._process_component(component_name)

    def process(self, app_names=None):
        if not app_names:
            app_names = self.requested_app_names

        for app_name in app_names:
            self._process_app(app_name)

        return self.k8s_list



import base64
import logging
import click
import json
import sys
import yaml

from bonfire.openshift import oc, wait_for_ready
from bonfire.utils import split_equals

from env_parser import EnvParser

SECRET_NAME = "iqe-settings"


def _get_base_pod_cfg():
    return {
        "apiVersion": "v1",
        "kind": "Pod",
        "metadata": {"name": "iqe-tests"},
        "spec": {
            "serviceAccountName": "iqe",
            "containers": [
                {
                    "command": ["/bin/cat"],
                    "image": "quay.io/cloudservices/iqe-tests:latest",
                    "imagePullPolicy": "Always",
                    "name": "iqe-tests",
                    "resources": {
                        "limits": {"cpu": "1", "memory": "2Gi"},
                        "requests": {"cpu": "500m", "memory": "1Gi"},
                    },
                    "stdin": True,
                    "tty": True,
                    "env": [{"name": "IQE_TESTS_LOCAL_CONF_PATH", "value": "/iqe_settings"}],
                    "volumeMounts": [{"mountPath": "/iqe_settings", "name": "iqe-settings-volume"}],
                }
            ],
            "imagePullSecrets": [{"name": "quay-cloudservices-pull"}],
            "restartPolicy": "Never",
            "volumes": [
                {
                    "name": "iqe-settings-volume",
                    "secret": {"defaultMode": 420, "secretName": SECRET_NAME},
                }
            ],
        },
    }


def _build_test_conf(env_parser):
    conf = {}
    env_name = "clowder_smoke"
    env_conf = conf[env_name] = {}

    # mq plugin configuration is now present in the plugin's settings.default.yaml

    # ingress configuration is now present in the plugin's settings.default.yaml

    # host-inventory configuration is now present in the plugin's settings.default.yaml

    if env_parser.app_present("marketplace"):
        mp_storage_cfg = env_parser.get_storage_config("marketplace")
        bucket = env_parser.get_bucket_config("marketplace", "marketplace-s3")
        env_conf["MARKETPLACE"] = {
            "aws_access_key_id": bucket.accessKey,
            "aws_secret_access_key": bucket.secretKey,
            "aws_s3_endpoint": f"{mp_storage_cfg.hostname}:{mp_storage_cfg.port}",
            "aws_s3_bucket": bucket.name,
            "aws_s3_secure": "false",
            "service_objects": {
                "api_v1": {
                    "config": {
                        "hostname": env_parser.get_hostname("ingress", "ingress-service"),
                        "port": env_parser.get_port("ingress", "ingress-service"),
                        "scheme": "http",
                    }
                }
            },
        }

    if env_parser.app_present("advisor"):
        env_conf["ADVISOR"] = {
            "kafka_dc_name": env_parser.get_kafka_hostname("advisor").split(".")[0],
            "db_dc_name": "advisor-db",
            "service_dc_name": "advisor-service",
            "api_dc_name": "advisor-api",
            "upload_dc_name": "ingress-service",
            "pup_dc_name": "puptoo-processor",
            "kafka_dc_port": env_parser.get_kafka_port("advisor"),
            "engine_results_topic": env_parser.get_kafka_topic(
                "advisor", "platform.engine.results"
            ),
            "inventory_events_topic": env_parser.get_kafka_topic(
                "advisor", "platform.legacy-bridge.events"
            ),
            "payload_tracker_topic": env_parser.get_kafka_topic(
                "advisor", "platform.payload-status"
            ),
            "kafka_hooks_topic": env_parser.get_kafka_topic("advisor", "hooks.outbox"),
            "db_hostname": env_parser.get_db_config("advisor").hostname,
            "db_database": env_parser.get_db_config("advisor").name,
            "db_username": env_parser.get_db_config("advisor").username,
            "db_password": env_parser.get_db_config("advisor").password,
            "db_port": env_parser.get_db_config("advisor").port,
            "service_objects": {
                "api": {
                    "config": {
                        "hostname": env_parser.get_hostname("advisor", "api"),
                        "port": env_parser.get_port("advisor", "api"),
                        "scheme": "http",
                    }
                }
            },
        }
        if env_parser.app_present("rbac"):
            hostname = env_parser.get_hostname("rbac", "service")
            port = env_parser.get_port("rbac", "service")
            env_conf["ADVISOR"][
                "rbac_url"
            ] = f"http://{hostname}:{port}/api/rbac/v1/access/?application=advisor"
        if env_parser.app_present("host-inventory"):
            env_conf["ADVISOR"]["egress_topic"] = (
                env_parser.get_kafka_topic("host-inventory", "platform.inventory.host-egress"),
            )

    if env_parser.app_present("playbook-dispatcher"):
        env_conf["RHC"] = {
            "kafka": {
                "playbook_validation_topic": env_parser.get_kafka_topic(
                    "playbook-dispatcher", "platform.upload.validation"
                ),
            },
            "service_objects": {
                "playbook_dispatcher_api_v1": {
                    "config": {
                        "hostname": env_parser.get_hostname(
                            "playbook-dispatcher", "playbook-dispatcher-api"
                        ),
                        "port": env_parser.get_port(
                            "playbook-dispatcher", "playbook-dispatcher-api"
                        ),
                        "scheme": "http",
                    }
                },
                "playbook_dispatcher_internal_api_v1": {
                    "config": {
                        "hostname": env_parser.get_hostname(
                            "playbook-dispatcher", "playbook-dispatcher-api"
                        ),
                        "port": env_parser.get_port(
                            "playbook-dispatcher", "playbook-dispatcher-api"
                        ),
                        "scheme": "http",
                    }
                },
            },
        }

    return conf


def _create_conf_secret(namespace):
    env_parser = EnvParser(namespace)
    conf_data = _build_test_conf(env_parser)
    encoded_conf = base64.b64encode(yaml.dump(conf_data).encode()).decode()
    secret = {
        "apiVersion": "v1",
        "kind": "Secret",
        "metadata": {"name": SECRET_NAME},
        "data": {"settings.local.yaml": encoded_conf},
    }
    oc("apply", f="-", n=namespace, _in=json.dumps(secret))


def _create_pod(namespace, pod_name, env):
    pod = _get_base_pod_cfg()

    pod["metadata"]["name"] = pod_name
    env_vars = split_equals(env, allow_null=True)
    if env_vars:
        pod_env_vars = pod["spec"]["containers"][0]["env"]
        for key, val in env_vars.items():
            if val:
                pod_env_vars.append({"name": key, "value": val})

    oc("apply", f="-", n=namespace, _in=json.dumps(pod))


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.argument("namespace", type=str, required=True)
@click.option("--pod-name", type=str, default="iqe-tests", help="name of pod (default: iqe-tests)")
@click.option(
    "--env",
    "-e",
    type=str,
    multiple=True,
    help="Env var to set on container using format KEY=VAL",
)
def main(namespace, pod_name, env):
    logging.basicConfig(level=logging.INFO)
    logging.getLogger("sh").setLevel(logging.CRITICAL)

    _create_conf_secret(namespace)
    _create_pod(namespace, pod_name, env)

    if not wait_for_ready(namespace, "pod", pod_name):
        sys.exit(1)
    print(pod_name)


if __name__ == "__main__":
    main()
